/* -*- c -*- */
/****************************************************************
 * number
 ****************************************************************/

Prim(numberp, "number?", 1)
{
  RETURN( SCM_MKBOOL(SCM_NUMBERP(TOS)) );
}

Prim(integerp, "integer?", 1)
{
  if (SCM_INUMP(TOS) || SCM_BNUMP(TOS)) { RETURN(scm_true); }
  if (SCM_FNUMP(TOS)) {
	double x = SCM_FNUM(TOS);
	RETURN(SCM_MKBOOL( floor(x) == x ));
  }
  RETURN(scm_false);
}

/* all numbers are real in this scheme */
Prim(realp, "real?", 1)
{
  RETURN(SCM_MKBOOL(SCM_NUMBERP(TOS)));
}

Prim(exactp, "exact?", 1)
{
  RETURN(SCM_MKBOOL( SCM_INUMP(TOS) || SCM_BNUMP(TOS) ));
}

Prim(inexactp, "inexact?", 1)
{
  RETURN(SCM_MKBOOL( SCM_FNUMP(TOS) ));
}

#ifdef INUM_OPTIMIZATION
#define OPTIM_LOGOP(op) \
if (SCM_INUMP((long)n1 & (long)TOS)) { \
  if ((long)n1 op (long)TOS) { TOS=scm_true; NEXT; } TOS=scm_false; NEXT; }
#else
#define OPTIM_LOGOP(op)
#endif

Prim(lt2, "*i-n2<*", 2)			/* n2 n1 -- flag */
{
  SOBJ n1;  spop(n1);  OPTIM_LOGOP(<);   TOS = scm_lt2(n1, TOS);  NEXT;
}

Prim(le2, "*i-n2<=*", 2)		/* n2 n1 -- flag */
{
  SOBJ n1;  spop(n1);  OPTIM_LOGOP(<=);  TOS = scm_le2(n1, TOS);  NEXT;
}

Prim(ge2, "*i-n2>=*", 2)		/* n2 n1 -- flag */
{
  SOBJ n1;  spop(n1);  OPTIM_LOGOP(>=);  TOS = scm_ge2(n1, TOS);  NEXT;
}

Prim(gt2, "*i-n2>*", 2)			/* n2 n1 -- flag */
{
  SOBJ n1;  spop(n1);  OPTIM_LOGOP(>);  TOS = scm_gt2(n1, TOS);  NEXT;
}

Prim(eq2, "*i-n2=*", 2)
{
  SOBJ n1;  spop(n1);  OPTIM_LOGOP(==);  TOS = scm_gt2(n1, TOS);  NEXT;
}

#define GEN_LOGOP(op) \
{ while((void*)(&sp[1]) < (void*)cont) { \
	if (!(scm_cmpnum(TOS,sp[1]) op 0)) { VRETURN(scm_false); } \
	sdrop(); \
  } \
  VRETURN(scm_true); \
}

PrimVarargs(ltv, "*i-nv<*")
{
  GEN_LOGOP(<);
}
PrimVarargs(lev, "*i-nv<=*")
{
  GEN_LOGOP(<=);
}
PrimVarargs(gev, "*i-nv>=*")
{
  GEN_LOGOP(>=);
}
PrimVarargs(gtv, "*i-nv>*")
{
  GEN_LOGOP(>);
}
PrimVarargs(eqv, "*i-nv=*")
{
  GEN_LOGOP(==);
}

Prim(zerop, "zero?", 1)
{
  if (SCM_INUMP(TOS)) {	RETURN(SCM_MKBOOL(SCM_INUM(TOS) == 0)); }
  RETURN(scm_zerop(TOS));
}

Prim(positivep, "positive?", 1)
{
  if (SCM_INUMP(TOS)) {	RETURN(SCM_MKBOOL(SCM_INUM(TOS) > 0)); }
  RETURN(scm_positivep(TOS));
}

Prim(negativep, "negative?", 1)
{
  if (SCM_INUMP(TOS)) {	RETURN(SCM_MKBOOL(SCM_INUM(TOS) < 0)); }
  RETURN(scm_negativep(TOS));
}

Prim(oddp, "odd?", 1)
{
  if (SCM_INUMP(TOS)) {	RETURN(SCM_MKBOOL( (SCM_INUM(TOS) & 1) == 1)); }
  RETURN(scm_oddp(TOS));
}

Prim(evenp, "even?", 1)
{
  if (SCM_INUMP(TOS)) {	RETURN(SCM_MKBOOL( (SCM_INUM(TOS) & 1) == 0)); }
  RETURN(scm_evenp(TOS));
}

PrimVarargs(min, "min")
{
  if (NARGS < 1) err("max: wrong number of args", NULL);
  sp++;
  if (SCM_INUMP(TOS)) {
	while( ((void *)sp < (void*)cont) && SCM_INUMP(*sp)) {
	  if (SCM_INUM(TOS) > SCM_INUM(*sp)) TOS = *sp;
	  sp++;
	}
  }
  while((void *)sp < (void*)cont) {
	if (scm_cmpnum(TOS, *sp) > 0) { TOS = *sp; }
	sp++;
  }
  VRETURN(TOS);
}

PrimVarargs(max, "max")
{
  if (NARGS < 1) err("max: wrong number of args", NULL);
  sp++;
  if (SCM_INUMP(TOS)) {
	while( ((void *)sp < (void*)cont) && SCM_INUMP(*sp)) {
	  if (SCM_INUM(TOS) < SCM_INUM(*sp)) TOS = *sp;
	  sp++;
	}
  }
  while((void *)sp < (void*)cont) {
	if (scm_cmpnum(TOS, *sp) < 0) { TOS = *sp; }
	sp++;
  }
  VRETURN(TOS);
}

Prim(add2, "add2", 2)				/* n2 n1 -- n1+n2 */
{
  SOBJ n1;
  spop(n1);

#ifdef INUM_OPTIMIZATION
  if (SCM_INUMP((long)n1 & (long)TOS)) {
	long r = SCM_INUM(n1) + SCM_INUM(TOS);
	if (SCM_INUM_RANGE(r)) {  TOS = SCM_MKINUM(r);  NEXT; }
	TOS = scm_int2bnum(r);  NEXT;
  }
#endif
  TOS = scm_add2(n1, TOS);
  NEXT;
}

PrimVarargs(addv, "addv")
{
  if ((void*)sp >= (void*)cont) { VRETURN(SCM_MKINUM(0)); }
  sp++;
  if (SCM_INUMP(TOS)) {
	long sum = SCM_INUM(TOS);
	long r = 0;
	while((void*)sp < (void*)cont && SCM_INUMP(*sp)) {
	  r = sum + SCM_INUM(*sp);
	  if (!SCM_INUM_RANGE(r)) 	break;
	  sum = r;
	  sp++;
	}
	TOS = SCM_MKINUM(sum);
  }
  while((void*)sp < (void*)cont) {
	TOS = scm_add2(TOS, *sp++);
  }
  sp--;
  VRETURN(TOS);
}

Prim(mul2, "mul2", 2)			/* n2 n1 -- n1*n2 */
{
  SOBJ n1;

  spop(n1);

#ifdef INUM_OPTIMIZATION
  if (SCM_INUMP(n1) && SCM_INUMP(TOS)) {
	long r, x, y;
	if ( (x = SCM_INUM(n1)) == 0 || (y = SCM_INUM(TOS)) == 0) {
	  TOS = SCM_MKINUM(0);
	  NEXT;
	}
	r = x * y;
	if (y == (r / x)) {
	  TOS = SCM_MKINUM(r);
	  NEXT;
	}
  }
#endif
  TOS = scm_mul2(n1, TOS);
  NEXT;
}


PrimVarargs(mulv, "mulv")
{
  if ((void*)sp >= (void*)cont) { VRETURN(SCM_MKINUM(1)); }
  sp++;
  if (SCM_INUMP(TOS)) {
	long sum = SCM_INUM(TOS);
	long r = 0;
	while((void*)sp < (void*)cont && SCM_INUMP(*sp)) {
	  r = sum * SCM_INUM(*sp);
	  if (sum != r / SCM_INUM(*sp)) break;
	  sum = r;
	  sp++;
	}
	TOS = SCM_MKINUM(sum);
  }
  while((void*)sp < (void*)cont) {
	TOS = scm_mul2(TOS, *sp++);
  }
  sp--;
  VRETURN(TOS);
}

Prim(sub2, "sub2", 2)		/* n2 n1 -- n1-n2 */
{
  SOBJ n1;

  spop(n1);
#ifdef INUM_OPTIMIZATION
  if (SCM_INUMP(n1) && SCM_INUMP(TOS)) {
	long r;
	r = SCM_INUM(n1) - SCM_INUM(TOS);
	if (SCM_INUM_RANGE(r)) {  TOS = SCM_MKINUM(r);  NEXT; }
	TOS = scm_int2bnum(r); NEXT;
  }
#endif
  TOS = scm_sub2(n1, TOS);
  NEXT;
}

PrimVarargs(subv, "subv")
{
  if ((void*)sp >= (void*)cont) { VRETURN(SCM_MKINUM(0)); }

  if (NARGS == 1) {	spush(SCM_MKINUM(0)); }
  sp++;
  if (SCM_INUMP(TOS)) {
	long sum = SCM_INUM(TOS);
	long r = 0;
	while((void*)sp < (void*)cont && SCM_INUMP(*sp)) {
	  r = sum - SCM_INUM(*sp);
	  if (!SCM_INUM_RANGE(r)) 	break;
	  sum = r;
	  sp++;
	}
	TOS = SCM_MKINUM(sum);
  }
  while((void*)sp < (void*)cont) {
	TOS = scm_sub2(TOS, *sp++);
  }
  VRETURN(TOS);
}

Prim(div2, "div2", 2)			/* n2 n1 -- n1/n2 */
{
  SOBJ n1;
  spop(n1);
  TOS = scm_div2(n1, TOS);
  NEXT;
}

PrimVarargs(divv, "divv")
{
  if (NARGS < 1) 	err("/: bad number of args", NULL);
  if (NARGS == 1)	spush(SCM_MKINUM(1));
  sp++;
  while((void*)sp < (void*)cont) {
	TOS = scm_div2(TOS, *sp++);
  }
  VRETURN(TOS);
}

Prim(abs, "abs", 1)
{
  if (SCM_INUMP(TOS)) {
	if (SCM_INUM(TOS) < 0) { TOS=SCM_MKINUM( -(SCM_INUM(TOS))); NEXT; }
  }
  RETURN(scm_abs(TOS));
}


Prim(quotient, "quotient", 2)
{
  SOBJ x;
  spop(x);
  RETURN(scm_quotient(x, TOS));
}

Prim(remainder, "remainder", 2)
{
  SOBJ x;
  spop(x);
  RETURN(scm_remainder(x, TOS));
}

Prim(modulo, "modulo", 2)
{
  SOBJ x;
  spop(x);
  RETURN(scm_modulo(x, TOS));
}

PrimVarargs(gcd, "gcd")
{
  supdate();  VRETURN(scm_gcd(NARGS, sp));
}

PrimVarargs(lcm, "lcm")
{
  supdate();  VRETURN(scm_lcm(NARGS, sp));
}

Prim(floor, 	"floor", 1)
{
  RETURN(scm_floor(TOS));
}
Prim(ceil, 		"ceiling", 1) 	
{
  RETURN(scm_ceil(TOS)); 
}
Prim(truncate, 	"truncate", 1) 	
{
  RETURN(scm_truncate(TOS)); 
}
Prim(round, 	"round", 1) 	
{
  RETURN(scm_round(TOS)); 
}

Prim(exp,		"exp",	1) 		
{
  RETURN(scm_exp(TOS)); 
}
Prim(log,		"log",	1) 		
{
  RETURN(scm_log(TOS)); 
}
Prim(sin,		"sin",	1) 		
{
  RETURN(scm_sin(TOS)); 
}
Prim(cos,		"cos",	1) 		
{
  RETURN(scm_cos(TOS)); 
}
Prim(tan,		"tan",	1) 		
{
  RETURN(scm_tan(TOS)); 
}
Prim(asin,		"asin",	1) 		
{
  RETURN(scm_asin(TOS)); 
}
Prim(acos,		"acos",	1) 		
{
  RETURN(scm_acos(TOS)); 
}

Prim(atan,		"atan",	2)
{
  SOBJ x;  spop(x);  RETURN(scm_atan(x, TOS));
}

Prim(sqrt,		"sqrt",	1)
{
  RETURN(scm_sqrt(TOS));
}

Prim(expt,		"expt",	2)
{
  SOBJ x;  spop(x);  RETURN(scm_expt(x, TOS));
}

Prim(exact2inexact, "exact->inexact", 1)
{
  RETURN(scm_exact_to_inexact(TOS));
}

Prim(inexact2exact, "inexact->exact", 1)
{
  RETURN(scm_inexact_to_exact(TOS));
}

PrimVarargs(number2string, "number->string")
{
  if (NARGS < 1) err("number->string: bad number of args", NULL);
  VRETURN(scm_number_to_string(TOS, (NARGS > 1) ? sp[1] : NULL));
}

PrimVarargs(string2number, "string->number")
{
  if (NARGS < 1) err("string->number: bad number of args", NULL);
  VRETURN(scm_string_to_number(TOS, (NARGS > 1) ? sp[1] : NULL));
}


Prim(plus1, "1+", 1)
{
  if (SCM_INUMP(TOS) && (SCM_INUM(TOS) < SOBJ_INUM_MAX)) {
	(long)TOS += (1 << SOBJ_INUM_SHIFT);
	NEXT;
  }
  TOS = scm_add2(SCM_MKINUM(1), TOS);
  NEXT;
}

Prim(plus2, "2+", 1)
{
  if (SCM_INUMP(TOS) && (SCM_INUM(TOS) < SOBJ_INUM_MAX)) {
	(long)TOS += (2 << SOBJ_INUM_SHIFT);
	NEXT;
  }
  TOS = scm_add2(SCM_MKINUM(2), TOS);
  NEXT;
}

Prim(minus1, "1-", 1)
{
  if (SCM_INUMP(TOS) && (SCM_INUM(TOS) > SOBJ_INUM_MIN)) {
	(long)TOS += (-1 << SOBJ_INUM_SHIFT);
	NEXT;
  }
  TOS = scm_sub2(TOS, SCM_MKINUM(1));
  NEXT;
}

Prim(minus2, "2-", 1)
{
  if (SCM_INUMP(TOS) && (SCM_INUM(TOS) > SOBJ_INUM_MIN)) {
	(long)TOS += (-2 << SOBJ_INUM_SHIFT);
	NEXT;
  }
  TOS = scm_sub2(TOS, SCM_MKINUM(2));
  NEXT;
}

