/*
 * Includes for port.c
 *
 * $Id: port.h 1.5 Wed, 06 Jan 1999 02:35:46 +0100 crad $
 */

typedef struct PORT {
  short type;
  short flag;
  int line;						/* line number starting from 0 */
  union {
	FILE *f;
	struct { int alloced; int index; int length; char *data;  } s;
  } descr;
} PORT;

#define PORT_R		1
#define PORT_W		2
#define PORT_RW		(PORT_R | PORT_W)

/*-- must be kept sync with port_driver[] */
#define PORT_T_FILE   0
#define PORT_T_STRING 1

#define PORT_MAX_TYPES	16		/* max number of port types */

/*-- special symbols */
#define PORT_EOF		EOF
#define PORT_STR_QTUM	128

/* driver depending on the port type */

typedef struct {
  PORT *(*open)		(void *p, int mode);
  void 	(*close)	(struct PORT *p);
  int 	(*getc)		(struct PORT *p);
  int 	(*peekc)	(struct PORT *p);
  void	(*putc)		(struct PORT *p, char c);
} PORT_DESCR;


/*-- prototypes */
PORT        *port_open(int type, void *p, int mode);
void        port_close(PORT *p);

int         port_getc(PORT *p);
int         port_peekc(PORT *p);
void        port_putc(PORT *p, char c);

PORT        *port_open_input_file(char *fname);
PORT        *port_open_output_file(char *fname);
PORT        *port_open_input_string(char *string);
PORT        *port_open_output_string();

void        port_puts(PORT *p, char *str);
void        port_putn(PORT *p, long n);
void        port_putx(PORT *p, void *ptr);
void        port_putd(PORT *p, double n);
