; -*- Scheme -*-
;
(print "; loading kernel")

(load "macro.scm")

;
; macro to speedup integer arith
;

(define-macro (+ . args)
  (let ((len (length args)))
;	(display "plus len=") (display len) (display " args=") (print args)
	(cond
	 ((eq? len 0) 	0)
	 ((eq? len 1)	(car args))
	 ((eq? len 2)
	  (let ((x (car args)) (y (cadr args)))
		(cond
		 ((eq? x 1)  `(1+ ,y))
		 ((eq? x 2)  `(2+ ,y))
		 ((eq? y 1)  `(1+ ,x))
		 ((eq? y 2)  `(2+ ,x))
		 (else `(add2 ,x ,y)))))
	  (else
	   `(addv ,@args)))))

(macro-set-func! + addv);

(define-macro (* . args)
  (let ((len (length args)))
	(cond
	 ((eq? len 0) 	1)
	 ((eq? len 1)	(car args))
	 ((eq? len 2)
	  (let ((x (car args)) (y (cadr args)))
		(cond
		 ((eq? x 1)  y)
		 ((eq? y 1)  x)
		 (else  `(mul2 ,x ,y)))))
	 (else `(mulv ,@args)))))

(macro-set-func! * mulv);

(define-macro (- . args)
  (let ((len (length args)))
;	(display "plus len=") (display len) (display " args=") (print args)
	(cond
	 ((eq? len 0) 	0)
	 ((eq? len 1)	`(sub2 0 ,(car args)))
	 ((eq? len 2)
	  (let ((x (car args)) (y (cadr args)))
		(cond
;		 ((eq? x 1)  `(1- ,y))
;		 ((eq? x 2)  `(2- ,y))
		 ((eq? y 1)  `(1- ,x))
		 ((eq? y 2)  `(2- ,x))
		 (else `(sub2 ,x ,y)))))
	  (else
	   `(subv ,@args)))))

(macro-set-func! - subv)

(define-macro (/ . args)
  (let ((len (length args)))
;	(display "plus len=") (display len) (display " args=") (print args)
	(cond
	 ((eq? len 0) 	0)
	 ((eq? len 1)	`(div2 1 ,(car args)))
	 ((eq? len 2)	`(div2 ,(car args) ,(cadr args)))
	 (else `(divv ,@args)))))

(macro-set-func! / divv)

;
; Math extensions
;
(define-macro (< . args)
  (let ((len (length args)))
	(cond ((or (eq? len 0) (eq? len 1)) #t)
		  ((eq? len 2) `(*i-n2<* ,(car args) ,(cadr args)))
		  (else `(*i-nv<* ,@args)))))

(macro-set-func! < *i-nv<*)

(define-macro (<= . args)
  (let ((len (length args)))
	(cond ((or (eq? len 0) (eq? len 1)) #t)
		  ((eq? len 2) `(*i-n2<=* ,(car args) ,(cadr args)))
		  (else `(*i-nv<=* ,@args)))))

(macro-set-func! <= *i-nv<=*)

(define-macro (>= . args)
  (let ((len (length args)))
	(cond ((or (eq? len 0) (eq? len 1)) #t)
		  ((eq? len 2) `(*i-n2>=* ,(car args) ,(cadr args)))
		  (else `(*i-nv>=* ,@args)))))

(macro-set-func! >= *i-nv>=*)

(define-macro (> . args)
  (let ((len (length args)))
	(cond ((or (eq? len 0) (eq? len 1)) #t)
		  ((eq? len 2) `(*i-n2>* ,(car args) ,(cadr args)))
		  (else `(*i-nv>* ,@args)))))

(macro-set-func! > *i-nv>*)

(define-macro (= . args)
  (let ((len (length args)))
	(cond ((or (eq? len 0) (eq? len 1)) #t)
		  ((eq? len 2) `(*i-n2=* ,(car args) ,(cadr args)))
		  (else `(*i-nv=* ,@args)))))

(macro-set-func! = *i-nv=*)
;
; Math constants
;
(define pi    3.14159265358979323846)
(define pi/2  1.57079632679489661923)
(define pi/4  0.78539816339744830962)


;
; FFI
;

;
; Note this define must match constants defined in dyn.c
;
; (define <any>			-1)
; (define <void>			0)
; (define <char>			1)
; (define <short>			2)
; (define <ushort>		3)
; (define <int>			4)
; (define <uint>			5)
; (define <long>			6)
; (define <ulong>			7)
; (define <float>			8)
; (define <double>		9)
; (define <static-ptr> 	10)
; (define <dynamic-ptr>	11)
; (define <string>		12)
; (define <boolean>		13)

; Example of FFI function
;
(define system  (make-extfunc "" :int "system"	'(:string)))

;
(define printf  (make-extfunc "" :void "printf"	'(:string . :any)))

(define fopen   (make-extfunc "" :item "fopen" '(:string :string)))

(define fclose  (make-extfunc "" :int "fclose" 	 '(:item)))

(define fprintf (make-extfunc "" :void "fprintf" '(:item . :any)))

; General purpose functions
;
(load "repl.scm")

